package goaml.lang.entity;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import goaml.lang.state.AbstractState;

public class Entity {

	private static final int DEFAULT_NUMBER_OF_INSTANCE = 1;
	private static final int DEFAULT_ID_LENGTH = 5;

	private final String NAME;
	private final String ID;

	// Contains every states of the object, even macro-states
	// This allows Process, Transition and Reference instances to find and get a state thanks to his
	// name with the method "getStateByName()"
	private final List<AbstractState> STATES;

	private Entity(final String name, final String id) {
		this.NAME = name;
		this.ID = id;
		this.STATES = new ArrayList<AbstractState>();
	}

	// From ONE "Entity" tag, create and return a list of corresponding instances of Entity
	public static List<Entity> parse(final Node node) {
		if (node == null) {
			throw new RuntimeException("Node is null");
		}
		if (node.getNodeName() != "Entity") {
			throw new RuntimeException("NodeName \"Entity\" was expected, got : " + node.getNodeName());
		}
		if (!node.hasChildNodes()) {
			throw new RuntimeException(
					"An entity node must contains others nodes (CompositeState, SequentialState and/or State)");
		}

		String name;
		int numberOfInstances = DEFAULT_NUMBER_OF_INSTANCE;
		int idLength = DEFAULT_ID_LENGTH;

		NamedNodeMap nodeMap = node.getAttributes();

		// Required attributes
		if (nodeMap.getNamedItem("name") != null) {
			name = nodeMap.getNamedItem("name").getTextContent();
		} else {
			throw new RuntimeException("\"name\" attribute missed");
		}

		// Optional attributes
		if (nodeMap.getNamedItem("numberOfInstances") != null) {
			if (Integer.valueOf(nodeMap.getNamedItem("numberOfInstances").getTextContent()) >= 0) {
				numberOfInstances = Integer.valueOf(nodeMap.getNamedItem("numberOfInstances").getTextContent());
			} else {
				throw new RuntimeException("\"numberOfInstances\" is negative in " + node.getNodeName() + " tag");
			}
		}

		if (nodeMap.getNamedItem("idLength") != null) {
			if (Integer.valueOf(nodeMap.getNamedItem("idLength").getTextContent()) >= 0) {
				idLength = Integer.valueOf(nodeMap.getNamedItem("idLength").getTextContent());
			} else {
				throw new RuntimeException("\"idLength\" is negative in " + node.getNodeName() + " tag");
			}
		}

		// "random" is created with a seed in order to have the same result every time for a given
		// modelization
		Random random = new Random(0);

		// List which contains every ID already generated
		List<String> ids = new ArrayList<String>();

		List<Entity> entities = new ArrayList<Entity>();

		// Create every instance of entity
		for (int numInstance = 0; numInstance < numberOfInstances; numInstance++) {
			// Generate an UNIQUE id of idLength size which contains only 0 and 1
			// making sure it has not already been generated
			StringBuilder id;
			do {
				id = new StringBuilder();
				for (int i = 0; i < idLength; i++) {
					id.append(random.nextInt(2));
				}
			} while (ids.contains(id.toString()));

			// We add the id to the list of unique id
			ids.add(id.toString());

			// Create a new entity
			Entity entity = new Entity(name, id.toString());

			// Create the entity's states
			for (int numState = 0; numState < node.getChildNodes().getLength(); numState++) {
				AbstractState.parse(entity, node.getChildNodes().item(numState), true);
			}

			entities.add(entity);
		}

		return entities;
	}

	public final String getName() {
		return this.NAME;
	}

	public final String getId() {
		return this.ID;
	}

	public final List<AbstractState> getStates() {
		return this.STATES;
	}

	public final List<AbstractState> getStatesByName(final String name) {
		List<AbstractState> results = new ArrayList<AbstractState>();

		for (AbstractState state : STATES) {
			if (state.getName().equals(name)) {
				results.add(state);
			}
		}

		return results;
	}

	public final void addState(final AbstractState state) {
		this.STATES.add(state);
	}

}
