package goaml.lang.process;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import goaml.lang.StateChart;
import goaml.lang.entity.Entity;
import goaml.lang.state.AbstractState;
import goaml.lang.state.State;

public abstract class AbstractProcess {

	protected AbstractProcess() {
	}

	// From ONE "AbstractProcess" tag, get attributes described in the tag
	// Then get the entities concerned by each Input or Output sub-tag and their state concerned
	// these inputs/outputs
	// Get also the values contain in "value" subNode
	// Then get the "real" process class concerned by the tag (obtains with "class" attribute)
	// and this process class will create every processes that will be added to the stateChart
	public static final void parse(final Node node, final StateChart stateChart) {

		if (node == null) {
			throw new RuntimeException("\"node\" is null");
		}
		if (node.getNodeName() != "Process") {
			throw new RuntimeException("NodeName \"Process\" was expected, got : " + node.getNodeName());
		}
		if (!node.hasChildNodes()) {
			throw new RuntimeException("A Process node must contains others nodes (parameters, etc.)");
		}

		// Inputs and outputs
		Hashtable<String, List<State>> inputsOuputs = new Hashtable<String, List<State>>();

		// Constants
		Hashtable<String, Double> values = new Hashtable<String, Double>();

		// Class
		String clazz;

		// =========================
		// Initialize attributes
		// =========================
		// Required
		NamedNodeMap nodeMap = node.getAttributes();
		if (nodeMap.getNamedItem("class") != null) {
			clazz = nodeMap.getNamedItem("class").getTextContent();
		} else {
			throw new RuntimeException("\"class\" attribute missed");
		}

		// For each subNodes
		for (int i = 0; i < node.getChildNodes().getLength(); i++) {
			Node subNode = node.getChildNodes().item(i);
			NamedNodeMap subNodeMap = subNode.getAttributes();

			switch (subNode.getNodeName()) {
			case "input":
			case "output":
				String name = null, entityName = null, stateName = null;
				int minCohortGuard = 0, maxCohortGuard = -1;

				// Required
				if (subNodeMap.getNamedItem("name") != null) {
					name = subNodeMap.getNamedItem("name").getTextContent();
				} else {
					throw new RuntimeException("\"name\" attribute missed in " + subNode.getNodeName() + " tag");
				}
				if (subNodeMap.getNamedItem("entity") != null) {
					entityName = subNodeMap.getNamedItem("entity").getTextContent();
				} else {
					throw new RuntimeException("\"entity\" attribute missed in " + subNode.getNodeName() + " tag");
				}
				if (subNodeMap.getNamedItem("state") != null) {
					stateName = subNodeMap.getNamedItem("state").getTextContent();
				} else {
					throw new RuntimeException("\"state\" attribute missed in " + subNode.getNodeName() + " tag");
				}

				// Optional
				if (subNodeMap.getNamedItem("minCohortGuard") != null) {
					if (Integer.valueOf(subNodeMap.getNamedItem("minCohortGuard").getTextContent()) >= 0) {
						minCohortGuard = Integer.valueOf(subNodeMap.getNamedItem("minCohortGuard").getTextContent());
					} else {
						throw new RuntimeException("minCohortGuard is negative in " + subNode.getNodeName() + " tag");
					}
				}
				if (subNodeMap.getNamedItem("maxCohortGuard") != null) {
					if (Integer.valueOf(subNodeMap.getNamedItem("maxCohortGuard").getTextContent()) >= 0) {
						maxCohortGuard = Integer.valueOf(subNodeMap.getNamedItem("maxCohortGuard").getTextContent());
					} else {
						throw new RuntimeException("maxCohortGuard is negative in " + subNode.getNodeName() + " tag");
					}
				}

				if (maxCohortGuard != -1 && minCohortGuard > maxCohortGuard) {
					throw new RuntimeException("minCohortGuard is greater than maxCohortGuard");
				}

				List<State> currentStates = new ArrayList<State>();

				List<Entity> entities = stateChart.getEntitiesByName(entityName);
				for (Entity entity : entities) {
					for (AbstractState state : entity.getStatesByName(stateName)) {
						if (state.getNumCohort() < minCohortGuard
								|| (maxCohortGuard != -1 && state.getNumCohort() > maxCohortGuard)) {
							continue;
						}
						if (state instanceof State) {
							currentStates.add((State) state);
						} else {
							throw new RuntimeException("Process inputs must be instance of \"State\"");
						}
					}
				}
				inputsOuputs.put(name, currentStates);

				break;

			case "value": // Optional
				String valueName = null;
				double valueRate = 0.0;

				if (subNodeMap.getNamedItem("name") != null) {
					valueName = subNodeMap.getNamedItem("name").getTextContent();
				} else {
					throw new RuntimeException("\"name\" attribute missed in " + subNode.getNodeName() + " tag");
				}
				if (subNodeMap.getNamedItem("value") != null) {
					if (Double.valueOf(subNodeMap.getNamedItem("value").getTextContent()) >= 0) {
						valueRate = Double.valueOf(subNodeMap.getNamedItem("value").getTextContent());
					} else {
						throw new RuntimeException(
								"\"value\" attribute is negative in " + subNode.getNodeName() + " tag");
					}
				} else {
					throw new RuntimeException("\"value\" attribute missed in " + subNode.getNodeName() + " tag");
				}

				values.put(valueName, valueRate);
				break;

			default:
				throw new RuntimeException("Invalid node tag : " + subNode.getNodeName());
			}
		}

		try {
			// Get the real process (subClass of Process)
			AbstractProcess process = (AbstractProcess) Class.forName("goaml.lang.process." + clazz).newInstance();

			// Call createProcess from subclasses which will create every real
			// transitions
			process.createProcess(stateChart, inputsOuputs, values);
		} catch (InstantiationException | IllegalAccessException | ClassNotFoundException e) {
			e.printStackTrace();
			throw new RuntimeException("Failed to create the transition !");
		}
	}

	protected abstract void createProcess(StateChart stateChart, Hashtable<String, List<State>> inputsOuputs,
			Hashtable<String, Double> values);

	// Create DelayedUpdates in order to apply computed value on the states when all transitions
	// and processes have compute
	public abstract void compute(final StateChart stateChart);

}
